/***********************************************************************\
*                              expression.d                             *
*                                                                       *
*                         Come Here interpreter                         *
*                                                                       *
*                             Version 1.0.1                             *
*                                                                       *
*                           Expression classes                          *
>***********************************************************************<
* Copyright (C) 2006-2007 Stewart Gordon.                               *
*                                                                       *
* This software is provided 'as-is', without any express or implied     *
* warranty.  In no event will the author be held liable for any damages *
* arising from the use of this software.                                *
*                                                                       *
* Permission is granted to anyone to use this software for any purpose, *
* including commercial applications, and to alter it and redistribute   *
* it freely, subject to the following restrictions:                     *
*                                                                       *
* - The origin of this software must not be misrepresented; you must    *
*   not claim that you wrote the original software.  If you use this    *
*   software in a product, an acknowledgment in the product             *
*   documentation would be appreciated but is not required.             *
* - Altered source versions must be plainly marked as such, and must    *
*   not be misrepresented as being the original software.               *
* - This notice may not be removed or altered from any source           *
*   distribution.                                                       *
*                                                                       *
* For documentation and other information related to this product,      *
* visit http://smjg.port5.com/wwwep/languages/comehere/.                *
* Email me with comments/suggestions/bug reports: smjg@iname.com        *
\***********************************************************************/
module smjg.prog.comehere.expression;

import smjg.prog.comehere.statement;
import smjg.prog.comehere.operator;
import smjg.prog.comehere.exception;
private {
	import std.string;
	import etc.bigint.bigint_int;
	debug import std.stdio;
}

/**
 *	General expression
 */
interface Expression {
	/**
	 *	Gets the value to which the expression currently evaluates.
	 */
	Int value();

	/**
	 *	Determines whether the expression depends on this variable.
	 */
	bool dependsOn(Variable v);

	char[] toString();
}


class Variable : Expression {
	char[] name;
	Int _value;

	ComeFromStatement[] comeFroms;

	this(char[] n) {
		name = n;
	}

	Int value() {
		debug {
			writefln("Entered Variable.value");
		}
		if (_value is null) {
			throw new UninitialisedVariableException(name);
		}

		return _value;
	}

	void value(Int v) {
		_value = v;
		foreach (ComeFromStatement cf; comeFroms) {
			cf.validate();
		}
	}

	char[] toString() {
		return name;
	}

	bool dependsOn(Variable v) { return v is this; }
}


Int numberFromString(char[] s) {
	void[] bytes = s;
	if (s.length % uint.sizeof != 0) {
		bytes.length = ((bytes.length / uint.sizeof) + 1)
		  * uint.sizeof;
	}
	uint[] uints = cast(uint[]) bytes;

	convertEndian(uints);

	return new Int(cast(uint[]) bytes, false);
}

version (LittleEndian) {
	void convertEndian(uint[] data) {}

} else {
	static assert (uint.sizeof == 4);
	void convertEndian(uint[] data) {
		foreach (inout uint d; data) {
			d = ((d & 0x000000FF) << 24) | ((d & 0x0000FF00) <<  8)
			  | ((d & 0x00FF0000) >>  8) | ((d & 0xFF000000) >> 24);
		}
	}
}


/**
 *	Wrapper for BigInteger.
 */
class NumberExpression : Expression {
	Int _value;

	this(Int v) {
		_value = v;
	}

	this(char[] str) {
		_value = numberFromString(str);
	}

	Int value() { return _value; }

	char[] toString() {
		return _value.toString;
	}

	bool dependsOn(Variable v) { return false; }
}


class ArithmeticExpression : Expression {
	Expression operand1;
	Operator operator;
	Expression operand2;

	this(Expression opd1, Operator op, Expression opd2) {
		operand1 = opd1;
		operator = op;
		operand2 = opd2;
	}

	Int value() {
		switch (operator.op) {
			case Operator.PLUS:
				return operand1.value + operand2.value;

			case Operator.MINUS:
				return operand1.value - operand2.value;

			case Operator.TIMES:
				return operand1.value * operand2.value;

			case Operator.DIVIDE:
				Int r;
				return divMod(operand1.value, operand2.value, r,
				  Int.Round.TOWARD_MINUS_INFINITY);

			case Operator.MOD:
				Int r;
				divMod(operand1.value, operand2.value, r,
				  Int.Round.TOWARD_MINUS_INFINITY);
				return r;
		}
	}

	public char[] toString() {
		return format("(%s %s %s)", operand1.toString, operator,
		  operand2.toString);
	}

	public bool dependsOn(Variable v) {
		return operand1.dependsOn(v) || operand2.dependsOn(v);
	}
}


class SgnExpression : Expression {
	Expression operand;

	this(Expression opd) {
		operand = opd;
	}

	Int value() {
		switch (sgn(operand.value)) {
			case -1:
				return Int.MINUS_ONE;
			case 0:
				return Int.ZERO;
			case 1:
				return Int.ONE;
		}
	}

	char[] toString() {
		return "SGN(" ~ operand.toString ~ ")";
	}

	bool dependsOn(Variable v) { return operand.dependsOn(v); }
}
