/***********************************************************************\
*                                parser.d                               *
*                                                                       *
*                         Come Here interpreter                         *
*                                                                       *
*                             Version 1.0.1                             *
*                                                                       *
*                           Parsing algorithm                           *
>***********************************************************************<
* Copyright (C) 2006-2007 Stewart Gordon.                               *
*                                                                       *
* This software is provided 'as-is', without any express or implied     *
* warranty.  In no event will the author be held liable for any damages *
* arising from the use of this software.                                *
*                                                                       *
* Permission is granted to anyone to use this software for any purpose, *
* including commercial applications, and to alter it and redistribute   *
* it freely, subject to the following restrictions:                     *
*                                                                       *
* - The origin of this software must not be misrepresented; you must    *
*   not claim that you wrote the original software.  If you use this    *
*   software in a product, an acknowledgment in the product             *
*   documentation would be appreciated but is not required.             *
* - Altered source versions must be plainly marked as such, and must    *
*   not be misrepresented as being the original software.               *
* - This notice may not be removed or altered from any source           *
*   distribution.                                                       *
*                                                                       *
* For documentation and other information related to this product,      *
* visit http://smjg.port5.com/wwwep/languages/comehere/.                *
* Email me with comments/suggestions/bug reports: smjg@iname.com        *
\***********************************************************************/
module smjg.prog.comehere.parser;

import smjg.prog.comehere.program;
import smjg.prog.comehere.expression;
import smjg.prog.comehere.operator;
import smjg.prog.comehere.exception;
private {
	import std.ctype;
	import etc.bigint.bigint_int;
	debug import std.stdio;
	debug (tokeniseResult) import std.stdio;
}


class Parser {
	Program program;

	Token[] tokens;
	Variable[char[]] variables;

	// used by parseExpression and its callees
	Expression[] operands;
	Operator[] operators;

	this(Program prog) {
		program = prog;
	}

	void parse(char[] code) {
		tokenise(code);
		debug writefln("Program tokenised");
		debug (tokeniseResult) foreach (Token t; tokens[0..$-1]) {
			writefln(t);
		}
		parse();
		debug writefln("Program parsed");
		program.labels.sort;
		debug writefln("Labels sorted");
	}

	void tokenise(char[] code) {
		// *** optimise to tokenise in-place rather than copying char-by-char
		char[] currentToken;
		Token lastToken = null;
		int currentTokenType = Token.WHITE;

		while (code.length != 0) {
			int tokenType;
			if (isspace(code[0])) {
				tokenType = Token.WHITE;
			} else if (isupper(code[0])) {
				tokenType = Token.KEYWORD;
			} else if (islower(code[0])) {
				tokenType = Token.IDENT;
			} else if (std.ctype.isdigit(code[0])) {
				tokenType = Token.NUMBER;
			} else if (code[0] == '"') {
				tokenType = Token.STRING;
			} else {
				tokenType = Token.SYMBOL;
			}

			// end of token?
			switch (currentTokenType) {
				case Token.WHITE:
					break;

				case Token.STRING:
					if (tokenType == Token.STRING) {	// end of string
						lastToken = new Token(Token.STRING, currentToken);
					}
					tokenType = Token.WHITE;
					break;

				case Token.SYMBOL:
					if (code[0] != '/' || currentToken != "/") {
						lastToken = new Token(Token.SYMBOL, currentToken);
					}
					break;

				default:
					if (tokenType != currentTokenType) {
						lastToken
						  = new Token(currentTokenType, currentToken);
					}
			}

			if (lastToken !is null) {
				tokens ~= lastToken;
				currentTokenType = Token.WHITE;
				currentToken.length = 0;
				lastToken = null;
			}

			if (currentTokenType == Token.WHITE) {
				if (tokenType != Token.WHITE) {	// begin new token
					if (tokenType != Token.STRING) {
						currentToken ~= code[0];
					}
					currentTokenType = tokenType;
				}
			} else {	// continue current token
				currentToken ~= code[0];
			}

			code = code[1..$];
		}
		tokens ~= null;
	}

	void parse() {
		//Token* tok = &tokens[0];
		Statement currentStatement = null;

		for (Token* tok = &tokens[0]; *tok !is null; tok++) {
			// get keyword or label
			Int label;
			int statement;

			if (tok.tokenType == Token.NUMBER) {	// labelled statement
				label = new Int(tok.data);
				// get actual statement keyword
				tok++;

				foreach (Int lbl; program.labels) {
					if (lbl == label) {
						throw new ProgramParseException(
						  "Duplicate label " ~ label.toString);
					}
				}
				program.labels ~= label;

			} else {
				label = null;
			}

			statement = tok.isStatementKeyword;

			// invalid statement
			if (tok.tokenType != Token.KEYWORD || statement < 0) {
				throw new ProgramParseException(
					"Expected statement keyword, but found "
					  ~ *tok);
			}

			switch (statement) {
				case Token.NOTE:	// comment - just skip tokens
					// at least 1 token in comment - extract it
					do {
						tok++;
					} while (!atNewStatement(tok));

					currentStatement
					  = new NullStatement(currentStatement, label);
					program.statements ~= currentStatement;

					break;

				case Token.ASK:
					tok++;

					if (tok.tokenType != Token.IDENT) {
						throw new ProgramParseException(
						  "Expected identifier, but found " ~ *tok);
					}

					currentStatement = new AskStatement(currentStatement,
					  label, getVariable(tok.data));
					program.statements ~= currentStatement;

					break;

				case Token.CALL:
					Expression rvalue = parseExpression(tok);
					tok++;
					if (tok.tokenType != Token.IDENT) {
						throw new ProgramParseException(
						  "Expected identifier, but found " ~ *tok);
					}

					currentStatement = new CallStatement(currentStatement,
						label, rvalue, getVariable(tok.data));
					program.statements ~= currentStatement;

					break;

				case Token.COME:
					tok++;
					if (tok.tokenType != Token.KEYWORD
						  || tok.data != "FROM") {
						throw new ProgramParseException(
							"Expected FROM, but found " ~ *tok);
					}
					Expression source = parseExpression(tok);

					currentStatement = new ComeFromStatement(
					  program, currentStatement, label, source);
					program.statements ~= currentStatement;
					program.allComeFroms
					  ~= cast(ComeFromStatement) currentStatement;

					// add to precedent variables
					foreach (Variable v; variables) {
						if (source.dependsOn(v)) {
							v.comeFroms
							  ~= cast(ComeFromStatement) currentStatement;
						}
					}
					break;

				case Token.TELL:
					Expression[] exps;

					do {
						exps ~= parseExpression(tok);
						debug (parseTell) {
							if (*tok is null) {
								writefln("Parsed expression, end of program");
							} else {
								writefln(
								  "Parsed expression, current token: %s",
								  *tok);
							}
						}
					} while (!atNewStatement(tok));

					currentStatement
					  = new TellStatement(currentStatement, label, exps);
					program.statements ~= currentStatement;
			}

			debug (parse) {
				assert (program.statements.length != 0);
				assert (program.statements[$-1] !is null);
				writefln(program.statements[$-1]);
			}
		}
	}

	/**
	 *	Check for beginning of new statement to terminate NOTE or TELL
	 */
	bool atNewStatement(Token* tok) {
		if (tok[1] is null) return true;

		bool labelled = false;

		if (tok[1].tokenType == Token.NUMBER) {
			if (tok[2] is null) return false;
			return tok[2].isStatementKeyword >= 0;
		} else {
			return tok[1].isStatementKeyword >= 0;
		}
	}

	Expression parseExpression(inout Token* tok) {
		operands.length = 0;
		operators.length = 0;
		tok++;

		bool expectingOperand = true, endOfExpression = false;

		while (*tok !is null && !endOfExpression) {
			debug (parseExpression) {
				writefln("Expecting opera%s, current token: %s",
				  expectingOperand ? "nd" : "tor", *tok);
			}

			if (expectingOperand) {
				switch (tok.tokenType) {
					case Token.NUMBER:
						operands ~= new NumberExpression(new Int(tok.data));
						expectingOperand = false;
						break;

					case Token.STRING:
						operands ~= new NumberExpression(tok.data);
						expectingOperand = false;
						break;

					case Token.IDENT:
						operands ~= getVariable(tok.data);
						expectingOperand = false;
						break;

					case Token.SYMBOL:
						operators ~= new Operator(tok.data, true);
						break;

					case Token.KEYWORD:
						if (tok.data == "SGN") {
							operators ~= new Operator(tok.data, true);

						} else {	// predefined constants
							// *** have ready-made Int objects
							if (tok.data == "BLANK") {
								operands ~= new NumberExpression(
								  new Int(cast(int) '\f'));
							} else if (tok.data == "NEXT") {
								operands ~= new NumberExpression(
								  new Int(cast(int) '\n'));
							} else if (tok.data == "QUOTE") {
								operands ~= new NumberExpression(
								  new Int(cast(int) '"'));
							} else {
								throw new ProgramParseException(
								  "Expected operand, but found " ~ *tok);
							}
							expectingOperand = false;
						}
						break;

					default:
						throw new ProgramParseException(
						  "Expected operand, but found " ~ *tok);
				}
				tok++;

			} else {	// expecting operator
				switch (tok.tokenType) {
					case Token.SYMBOL:
					case Token.KEYWORD:
						Operator newOp;
						try {
							newOp = new Operator(tok.data, false);

						} catch (ProgramParseException e) {
							// not valid operator - end of expression
							endOfExpression = true;
							debug (parseExpression) {
								writefln("Keyword ends expression");
							}
							break;
						}

						if (newOp.op == Operator.BRACKET) {
							// close bracket - dig down to the open bracket
							while (true) {
								if (operators.length == 0) {
									throw new ProgramParseException(
									  "Unopened close bracket");
								}
								if (operators[$-1].op == Operator.BRACKET) {
									break;
								} else {
									popExpressionPart();
								}
							}

							// now remove the open bracket
							operators.length = operators.length - 1;
							tok++;

						} else if (operators.length == 0 ||
							  (operators[$-1].stackBelow(newOp))) {
							// tighter binding - push it on
							operators ~= newOp;
							expectingOperand = true;
							tok++;
						} else {
							// looser binding - pop stuff
							popExpressionPart();
						}

						break;

					default:
						// end of expression
						endOfExpression = true;
						debug (parseExpression) {
							writefln("Non-keyword ends expression");
						}
				}
			}
		}

		tok--;

		// reached end of expression - finalise representation
		while (operators.length != 0) {
			popExpressionPart();
		}
		assert (operands.length == 1);
		return operands[0];
	}

	Variable getVariable(char[] name) {
		Variable* v = name in variables;

		if (v is null) {
			Variable var = new Variable(name);
			variables[name] = var;
			program.variables ~= var;
			return var;
		} else {
			return *v;
		}

		/+Variable var = variables[name];

		if (var is null) {
			var = new Variable(name);
			variables[name] = var;
			program.variables ~= var;
		}

		return var;+/
	}

	void popExpressionPart() {
		Operator op = operators[$-1];
		operators.length = operators.length - 1;
		if (op.op == Operator.BRACKET) {
			// reached end of expression, but brackets aren't balanced
			throw new ProgramParseException("Unterminated bracket");
		} else if (op.op == Operator.SGN) {
			operands[$-1] = new SgnExpression(operands[$-1]);
		} else {
			Expression op2 = operands[$-1];
			operands[$-2] = new ArithmeticExpression(operands[$-2], op, op2);
			operands.length = operands.length - 1;
		}
	}
}


class Token {
	enum : int {
		WHITE,
		KEYWORD,
		IDENT,
		NUMBER,
		STRING,
		SYMBOL
	}
	enum : int {
		ASK,
		CALL,
		COME,
		NOTE,
		TELL
	}

	int tokenType;
	char[] data;

	this(int tt, char[] d) {
		tokenType = tt;
		data = d.dup;
	}

	char[] toString() {
		if (tokenType == STRING) {
			return \" ~ data ~ \";
		} else {
			return data;
		}
	}

	int isStatementKeyword() {
		static const char[][] statementKeywords =
		  [ "ASK", "CALL", "COME", "NOTE", "TELL" ];

		if (tokenType == KEYWORD) {
			foreach (int index, char[] keyword; statementKeywords) {
				if (keyword == data) return index;
			}
		}
		return -1;
	}

	char[] opCat_r(char[] s) {
		return s ~ toString;
	}
}
