/***********************************************************************\
*                              statement.d                              *
*                                                                       *
*                         Come Here interpreter                         *
*                                                                       *
*                             Version 1.0.1                             *
*                                                                       *
*                           Statement classes                           *
>***********************************************************************<
* Copyright (C) 2006-2007 Stewart Gordon.                               *
*                                                                       *
* This software is provided 'as-is', without any express or implied     *
* warranty.  In no event will the author be held liable for any damages *
* arising from the use of this software.                                *
*                                                                       *
* Permission is granted to anyone to use this software for any purpose, *
* including commercial applications, and to alter it and redistribute   *
* it freely, subject to the following restrictions:                     *
*                                                                       *
* - The origin of this software must not be misrepresented; you must    *
*   not claim that you wrote the original software.  If you use this    *
*   software in a product, an acknowledgment in the product             *
*   documentation would be appreciated but is not required.             *
* - Altered source versions must be plainly marked as such, and must    *
*   not be misrepresented as being the original software.               *
* - This notice may not be removed or altered from any source           *
*   distribution.                                                       *
*                                                                       *
* For documentation and other information related to this product,      *
* visit http://smjg.port5.com/wwwep/languages/comehere/.                *
* Email me with comments/suggestions/bug reports: smjg@iname.com        *
\***********************************************************************/
module smjg.prog.comehere.statement;

import smjg.prog.comehere.program;
import smjg.prog.comehere.expression;
import smjg.prog.comehere.exception;
private {
	import std.string;
	import std.stream;
	import std.cstream;
	import smjg.api.c.disp;
	import etc.bigint.bigint_int;
}

/**
 *	Abstract statement class.
 */
abstract class Statement {
	Int label;
	ComeFromStatement[] comeFroms;
	Statement next;

	this(Statement prev, Int lbl) {
		if (prev !is null) prev.next = this;
		label = lbl;
	}

	abstract void execute();
	abstract char[] stringRep();

	final char[] toString() {
		if (label is null) {
			return stringRep;
		} else {
			return format("%s %s", label, stringRep);
		}
	}
}


/**
 *	Null statement - NOTE
 */
class NullStatement : Statement {
	this(Statement prev, Int lbl) { super(prev, lbl); }

	final public void execute() {}

	char[] stringRep() { return "NOTE."; }
}


/**
 *	COME FROM statement
 */
class ComeFromStatement : NullStatement {
	Expression source;

	// needed for this particular statement
	Program program;

	this(Program prog, Statement prev, Int lbl, Expression src) {
		super(prev, lbl);
		source = src;
		program = prog;
	}

	char[] stringRep() {
		return format("COME FROM %s", source.toString);
	}

	void validate() {
		try {
			Int value = source.value;
			// do binary search
			int lowUp = 0, highDown = program.labels.length;
			while (lowUp < highDown) {
				int midpoint = (lowUp + highDown) / 2;
				int cmp = value.opCmp(program.labels[midpoint]);

				if (cmp == 0) {
					return;
				} else if (cmp < 0) {
					highDown = midpoint;
				} else {
					lowUp = midpoint + 1;
				}
			}

			throw new ProgramRuntimeException(
			  "No statement with label " ~ value.toString);

		} catch (UninitialisedVariableException e) {
			if (true) {	// to avert DMD bug
				debug (trace) {
					writefln("COME FROM not yet in use");
				}
			}
		}
	}
}


/**
 *	ASK statement
 */
class AskStatement : Statement {
	Variable target;

	this(Statement prev, Int lbl, Variable trg) {
		super(prev, lbl);
		target = trg;
	}

	void execute() {
		target.value = numberFromString(din.readLine());
	}

	char[] stringRep() {
		return "ASK " ~ target.toString();
	}
}


/**
 *	TELL statement
 */
class TellStatement : Statement {
	Expression[] exps;

	static {
		char[] newLines(uint count) {
			char[] result = new char[count];
			result[] = '\n';
			return result;
		}

		const char[] ANSI = "\33[2J";

		version (Windows) {
			/+bool useDisp = true;
			char[] clearData;+/
			bool useDisp = false;
			char[] clearData = ANSI;
		} else version (linux) {
			bool useDisp = false;
			char[] clearData = ANSI;
		} else {
			bool useDisp = false;
			char[] clearData = "\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n"
							   "\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n"
							   "\n\n\n\n\n\n\n\n\n\n";
		}

		void clearScreen() {
			if (useDisp) {
				disp_move(0, 0);
				disp_eeop();
			} else {
				dout.writeString(clearData);
			}
		}
	}

	this(Statement prev, Int lbl, Expression[] es) {
		super(prev, lbl);
		exps = es;
	}

	public void execute() {
		foreach (Expression e; exps) {
			Int value = e.value;
			if (!value.equalsZero) {
				uint[] uints = value[0..value.end];

				convertEndian(uints);

				char[] text = cast(char[]) cast(void[]) uints;
				while (text[$-1] == '\0') text.length = text.length - 1;

				char[][] pages = split(text, "\f");
				foreach (char[] page; pages[0..$-1]) {
					dout.writeString(page);
					clearScreen();
				}
				dout.writeString(pages[$-1]);
			}
		}
	}

	char[] stringRep() {
		char[] buf = "TELL";
		foreach (Expression exp; exps) {
			buf ~= " " ~ exp.toString;
		}
		return buf;
	}
}


/**
 *	CALL statement
 */
class CallStatement : Statement {
	Expression exp;
	Variable target;

	this(Statement prev, Int lbl, Expression e, Variable trg) {
		super(prev, lbl);
		exp = e;
		target = trg;
	}

	void execute() {
		target.value = exp.value;
	}

	char[] stringRep() {
		return format("CALL %s %s", exp.toString, target);
	}
}
